import json
from functools import partial
from typing import Any, Sequence, Dict, Optional

from pylumi.ext import _pylumi


class Provider:
    """
    A pulumi provider logically maps to a real-world service or API, and in Pulumi
    terms maps to a resource provider process running locally that Pulumi communicates
    with via a gRPC interface. Common examples would be AWS or GCP.
    """

    def __init__(
        self, ctx: "Context", name: str, config: Optional[Dict[str, Any]] = None
    ) -> None:
        if config is None:
            config = {}

        self.name = name
        self.ctx = ctx
        self.config = config

    def configure(self, inputs: Optional[Dict[str, Any]] = None) -> None:
        """
        Configure this provider with the given configuration.

        **Parameters:**

        * **inputs** - (optional) configure this provider with the given configuration
        instead of the one passed in the constructor.

        **Returns:**

        None

        **Pulumi Docs:**

        Configure configures the resource provider with "globals" that control its behavior.

        Reference: `Configure <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        if inputs is None:
            inputs = self.config
        _pylumi.provider_configure(self.ctx.name, self.name, inputs)

    def teardown(self) -> None:
        """
        Tear down resources associated with this provider.

        **Returns:**

        None
        """
        _pylumi.provider_teardown(self.ctx.name, self.name)

    def get_schema(self, version: int = 0, decode: bool = True) -> Dict[str, Any]:
        """
        Get the schema information about this provider.

        **Parameters:**

        * **version** - (optional) specify a schema version for the provider. Default is 0.
        * **decode** - (optional) decode the raw JSON schema string and return a Python dictionary,
        defaluts to True.

        **Returns:**

        A Python dictionary with the decoded JSON schema information if `decode=True`. Otherwise,
        a bytes object that can be decoded via `json.loads()`.

        **Pulumi Docs:**

        Reference: `GetSchema <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        res = _pylumi.provider_get_schema(self.ctx.name, self.name, version)
        return json.loads(res) if decode else res

    def check_config(
        self,
        urn: str,
        olds: Dict[str, Any],
        news: Dict[str, Any],
        allow_unknowns: bool = False,
    ) -> Dict[str, Any]:
        """
        Validate the given provider configuration.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **olds** - old bag of properties
        * **news** - new bag of properties
        * **allow_unknowns** - (optional) allow unknown values in the output, default False.

        **Returns:**

        (properties, errors) tuple, where `properties` is a validated version of the configuration that
        should be passed to configure() and `errors` is a list of errors indicating validation errors or `None`.

        **Pulumi Docs:**

        CheckConfig validates the configuration for this resource provider.

        Reference: `CheckConfig <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_check_config(
            self.ctx.name, self.name, str(urn), olds, news, allow_unknowns
        )

    def diff_config(
        self,
        urn: str,
        olds: Dict[str, Any],
        news: Dict[str, Any],
        allow_unknowns: bool = False,
        ignore_changes: Sequence[str] = (),
    ) -> Dict[str, Any]:
        """
        Diff the given provider configurations.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **olds** - old bag of properties
        * **news** - new bag of properties
        * **allow_unknowns** - (optional) allow unknown values in the output, default False.

        **Returns:**

        A dictionary response containing information about the diff.

        **Pulumi Docs:**

        DiffConfig checks what impacts a hypothetical change to this provider's configuration will have on the provider.

        Reference: `DiffConfig <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_diff_config(
            self.ctx.name,
            self.name,
            str(urn),
            olds,
            news,
            allow_unknowns,
            ignore_changes,
        )

    def check(
        self,
        urn: str,
        olds: Dict[str, Any],
        news: Dict[str, Any],
        allow_unknowns: bool = False,
    ) -> Dict[str, Any]:
        """
        Validate the given resource configuration.

        **Parameters**

        * **urn** - pulumi resource URN.
        * **olds** - old bag of properties
        * **news** - new bag of properties
        * **allow_unknowns** - (optional) allow unknown values in the output, default False.

        **Returns:**
        (properties, errors) tuple, where `properties` is the validated bag of properties to be used
        for subsequent operations and errors is a list of validation errors, or None.

        **Pulumi Docs:**

        Check validates that the given property bag is valid for a resource of the given type and returns
        the inputs that should be passed to successive calls to Diff, Create, or Update for this resource.

        Reference: `Check <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_check(
            self.ctx.name, self.name, str(urn), olds, news, allow_unknowns
        )

    def diff(
        self,
        urn: str,
        id: str,
        olds: Dict[str, Any],
        news: Dict[str, Any],
        allow_unknowns: bool = False,
        ignore_changes: Sequence[str] = (),
    ) -> Dict[str, Any]:
        """
        Diff the given resource configurations.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **id** - pulumi resource ID.
        * **olds** - old bag of properties
        * **news** - new bag of properties
        * **allow_unknowns** - (optional) allow unknown values in the output, default False.

        **Returns:**

        A dictionary response containing information about the diff.

        **Pulumi Docs:**

        Diff checks what impacts a hypothetical update will have on the resource's properties.

        Reference: `Diff <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_diff(
            self.ctx.name,
            self.name,
            str(urn),
            id,
            olds,
            news,
            allow_unknowns,
            ignore_changes,
        )

    def create(
        self, urn: str, news: Dict[str, Any], timeout: int = 60, preview: bool = False
    ) -> Dict[str, Any]:
        """
        Create a pulumi resource.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **id** - pulumi resource ID.
        * **news** - new bag of properties
        * **timeout** - (optional) timeout for the operation, default 60
        * **preview** - (optional) predict the future state of the resource, default False.

        **Returns:**

        A dictionary with the following keys:

        * **ID** - The ID of the new created resource
        * **Properties** - A dictonary of properties of the new created resource.
        * **Status** - An integer status code for the operation

        **Pulumi Docs:**

        Create allocates a new instance of the provided resource and returns its unique resource.ID.

        Reference: `Create <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_create(
            self.ctx.name, self.name, str(urn), news, timeout, preview
        )

    def read(
        self, urn: str, id: str, inputs: Dict[str, Any], state: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Read the state of a pulumi resource.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **id** - plumi resource ID.
        * **inputs** - input properties
        * **state** - properties from the current state of the resource

        **Returns:**

        A dictionary with the following keys:

        * **ID** - The ID of the read resource.
        * **Inputs** - The dictionary of inputs for the read resource.
        * **Outputs** - The dictionary of outputs for the read resource.
        * **Status** - An integer status code from the operation.

        **Pulumi Docs:**

        Read the current live state associated with a resource.  Enough state must be include in the
        inputs to uniquely identify the resource; this is typically just the resource ID, but may also
        include some properties.  If the resource is missing (for instance, because it has been deleted),
        the resulting property map will be nil.

        Reference: `Read <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_read(
            self.ctx.name, self.name, str(urn), id, inputs, state
        )

    def update(
        self,
        urn: str,
        id: str,
        olds: Dict[str, Any],
        news: Dict[str, Any],
        timeout: int = 60,
    ) -> Dict[str, Any]:
        """
        Update the state of a pulumi resource.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **id** - pulumi resource ID.
        * **olds** - old bag of properties.
        * **news** - new bag of properties.
        * **timeout** - timeout for the operation, default 60.

        **Returns:**

        A dictionary with the following keys:

        * **ID** - The ID of the new created resource
        * **Properties** - A dictonary of properties of the new created resource.
        * **Status** - An integer status code for the operation

        **Pulumi Docs:**

        Update updates an existing resource with new values.

        Reference: `Update <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_update(
            self.ctx.name, self.name, str(urn), id, olds, news, timeout
        )

    def delete(self, urn: str, id: str, news: Dict[str, Any], timeout: int = 60) -> int:
        """
        Delete a pulumi resource.

        **Parameters:**

        * **urn** - pulumi resource URN.
        * **id** - pulumi resource ID.
        * **news** - new bag of properties.
        * **timeout** - timeout for the operation, default 60.

        **Returns:**

        An integer status code.

        **Pulumi Docs:**

        Delete tears down an existing resource.

        Reference: `Delete <https://github.com/pulumi/pulumi/sdk/v2/go/common/resource/provider.go>`_
        """
        return _pylumi.provider_delete(
            self.ctx.name, self.name, str(urn), id, news, timeout
        )

    def __enter__(self) -> Any:
        self.configure()
        return self

    def __exit__(self, exc_type, exc_value, tb) -> None:
        self.teardown()



# Modified from https://github.com/asottile/setuptools-golang-examples
import argparse
import contextlib
import copy
import errno
import os
import platform
import shlex
import shutil
import stat
import subprocess
import sys
import tempfile
from distutils.ccompiler import CCompiler
from distutils.dist import Distribution
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type

from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext


def rmtree(path: str) -> None:
    """Newer golang uses readonly dirs & files for module cache."""
    def handle_remove_readonly(
            func: Callable[..., Any],
            path: str,
            exc: Tuple[Type[OSError], OSError, TracebackType],
    ) -> None:
        excvalue = exc[1]
        if (
                func in (os.rmdir, os.remove, os.unlink) and
                excvalue.errno == errno.EACCES
        ):
            for p in (path, os.path.dirname(path)):
                os.chmod(p, os.stat(p).st_mode | stat.S_IWUSR)
            func(path)
        else:
            raise
    shutil.rmtree(path, ignore_errors=False, onerror=handle_remove_readonly)


@contextlib.contextmanager
def _tmpdir() -> Generator[str, None, None]:
    tempdir = tempfile.mkdtemp()
    try:
        yield tempdir
    finally:
        rmtree(tempdir)


def _get_cflags(
        compiler: CCompiler,
        macros: Sequence[Tuple[str, Optional[str]]],
) -> str:
    # https://github.com/python/typeshed/pull/3741
    args = [f'-I{p}' for p in compiler.include_dirs]  # type: ignore
    for macro_name, macro_value in macros:
        if macro_value is None:
            args.append(f'-D{macro_name}')
        else:
            args.append(f'-D{macro_name}={macro_value}')
    return ' '.join(args)


LFLAG_CLANG = '-Wl,-undefined,dynamic_lookup'
LFLAG_GCC = '-Wl,--unresolved-symbols=ignore-all'
LFLAGS = (LFLAG_CLANG, LFLAG_GCC)


def _get_ldflags() -> str:
    """Determine the correct link flags.  This attempts dummy compiles similar
    to how autotools does feature detection.
    """
    # windows gcc does not support linking with unresolved symbols
    if sys.platform == 'win32':  # pragma: no cover (windows)
        prefix = getattr(sys, 'real_prefix', sys.prefix)
        libs = os.path.join(prefix, 'libs')
        return '-L{} -lpython{}{}'.format(libs, *sys.version_info[:2])

    cc = subprocess.check_output(('go', 'env', 'CC')).decode('UTF-8').strip()

    with _tmpdir() as tmpdir:
        testf = os.path.join(tmpdir, 'test.c')
        with open(testf, 'w') as f:
            f.write('int f(int); int main(void) { return f(0); }\n')

        for lflag in LFLAGS:  # pragma: no cover (platform specific)
            try:
                subprocess.check_call((cc, testf, lflag), cwd=tmpdir)
                return lflag
            except subprocess.CalledProcessError:
                pass
        else:  # pragma: no cover (platform specific)
            # wellp, none of them worked, fall back to gcc and they'll get a
            # hopefully reasonable error message
            return LFLAG_GCC


def _check_call(cmd: Tuple[str, ...], cwd: str, env: Dict[str, str]) -> None:
    envparts = [f'{k}={shlex.quote(v)}' for k, v in sorted(tuple(env.items()))]
    print(
        '$ {}'.format(' '.join(envparts + [shlex.quote(p) for p in cmd])),
        file=sys.stderr,
    )
    subprocess.check_call(cmd, cwd=cwd, env=dict(os.environ, **env))


def _get_build_extension_methods(
        base: Type[_build_ext],
        root: str,
) -> Callable[[_build_ext, Extension], None]:

    def get_ext_filename(self, name: str) -> str:
        if name == 'libpylumigo':
            return 'libpylumigo.so'
        return base.get_ext_filename(self, name)

    def build_extension(self: _build_ext, ext: Extension) -> None:
        # If there are no .go files then the parent should handle this
        if not any(source.endswith('.go') for source in ext.sources):
            # the base class may mutate `self.compiler`
            compiler = copy.deepcopy(self.compiler)
            self.compiler, compiler = compiler, self.compiler

            ext_copy = copy.deepcopy(ext)
            build_lib = os.path.abspath(self.build_lib)
            ext_copy.include_dirs.append(build_lib)
            ext_copy.library_dirs.append('$ORIGIN')
            ext_copy.runtime_library_dirs.append('$ORIGIN')
            ext_copy.library_dirs.append(build_lib)
            ext_copy.runtime_library_dirs.append(build_lib)
            ext_copy.extra_link_args.append(f'-Wl,-rpath,{build_lib}')
            ext_copy.extra_link_args.append(f'-Wl,-rpath,$ORIGIN')

            ext_copy.libraries.append('pylumigo')

            try:
                return base.build_extension(self, ext_copy)
            finally:
                self.compiler, compiler = compiler, self.compiler

        if len(ext.sources) != 1:
            raise OSError(
                f'Error building extension `{ext.name}`: '
                f'sources must be a single file in the `main` package.\n'
                f'Recieved: {ext.sources!r}',
            )

        main_file, = ext.sources
        if not os.path.exists(main_file):
            raise OSError(
                f'Error building extension `{ext.name}`: '
                f'{main_file} does not exist',
            )
        main_dir = os.path.dirname(main_file)

        # Copy the package into a temporary GOPATH environment
        with _tmpdir() as tempdir:
            root_path = os.path.join(tempdir, 'src', root)
            # Make everything but the last directory (copytree interface)
            os.makedirs(os.path.dirname(root_path))
            shutil.copytree('.', root_path, symlinks=True)
            pkg_path = os.path.join(root_path, main_dir)

            env = {'GOPATH': tempdir}
            cmd_get = ('go', 'get', '-d')
            _check_call(cmd_get, cwd=pkg_path, env=env)

            env.update({
                'CGO_CFLAGS': _get_cflags(
                    self.compiler, ext.define_macros or (),
                ),
                'CGO_LDFLAGS': _get_ldflags(),
            })

            fname = ext.name + '.so'
            fpath = os.path.abspath(os.path.join(self.build_lib, fname))

            cmd_build = (
                'go', 'build', '-buildmode=c-shared',
                '-o', fpath,
            )
            _check_call(cmd_build, cwd=pkg_path, env=env)
            if platform.system() == 'Darwin':
                _check_call(['install_name_tool', '-id', '@loader_path/' + fname, fpath], cwd=pkg_path, env=env)

    return {
        'build_extension': build_extension,
        'get_ext_filename': get_ext_filename
    }


def _get_build_ext_cls(base: Type[_build_ext], root: str) -> Type[_build_ext]:
    return type('build_ext', (base,), _get_build_extension_methods(base, root))


def set_build_ext(
        dist: Distribution,
        attr: str,
        value: Dict[str, str],
) -> None:
    assert False
    root = value['root']
    # https://github.com/python/typeshed/pull/3742
    base = dist.cmdclass.get('build_ext', _build_ext)  # type: ignore
    dist.cmdclass['build_ext'] = _get_build_ext_cls(base, root)  # type: ignore



from pylumi import exc
from pylumi.context import Context
from pylumi.ext import (
    UNKNOWN_KEY,
    UNKNOWN_BOOL_VALUE,
    UNKNOWN_NUMBER_VALUE,
    UNKNOWN_ARRAY_VALUE,
    UNKNOWN_ASSET_VALUE,
    UNKNOWN_ARCHIVE_VALUE,
    UNKNOWN_OBJECT_VALUE,
)
from pylumi.provider import Provider
from pylumi.urn import URN

__version__ = "1.1.5"

__pulumi_version__ = "2.12.0"



import os
import platform

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

from setup_helper import _get_build_ext_cls

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

extensions = [Extension('libpylumigo', ['go/main.go'])]

try:
    from Cython.Build import cythonize
except ImportError:
    extensions.append(Extension('_pylumi', ['_pylumi.c']))
else:
    extensions.append(Extension('_pylumi', ['_pylumi.pyx']))
    extensions = cythonize(extensions, language_level='3str')

with open(os.path.join(CURRENT_DIR, 'README.rst')) as f:
    long_description = f.read().strip()

with open(os.path.join(CURRENT_DIR, 'requirements.txt')) as f:
    install_requires = list(filter(None, map(str.strip, f)))

with open(os.path.join(CURRENT_DIR, 'requirements-tests.txt')) as f:
    install_requires_tests = list(filter(None, map(str.strip, f)))

with open(os.path.join(CURRENT_DIR, 'requirements-dev.txt')) as f:
    install_requires_dev = list(filter(None, map(str.strip, f)))

setup(
    name='pylumi',
    version='1.1.5',
    description='Python API for interacting with Pulumi resource plugins.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Cython',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    keywords='go cython terraform pulumi infra-as-code cloudformation',
    url='https://github.com/cfeenstra67/pylumi',
    author='Cam Feenstra',
    author_email='cameron.l.feenstra@gmail.com',
    install_requires=install_requires,
    license='MIT',
    # Actual package data
    cmdclass={'build_ext': _get_build_ext_cls(_build_ext, 'github.com/cfeenstra67/pylumi')},
    packages=['pylumi'],
    py_modules=['setup_helper'],
    ext_modules=extensions,
    extras_require={
        'tests': install_requires_tests,
        'dev': install_requires_dev
    },
    package_data={
        '': [
            'requirements.txt',
            'requirements-dev.txt',
            'requirements-tests.txt',
            'README.md',
            'go/*'
        ],
    },
    include_package_data=True
)
